from wpilib import ADIS16448_IMU


class GyroADIS16448:
    def __init__(self) -> None:
        self._gyro = ADIS16448_IMU()
        self.__offset = 0

    @property
    def angle(self) -> float:
        return self._gyro.getGyroAngleZ() + self.__offset

    def reset(self):
        self.__offset = self.angle



class Motor:
    def init(self): ...
    def set_raw_output(self, x: float): ...


class EncoderMotor(Motor):
    def get_sensor_position(self) -> float: ...
    def get_sensor_velocity(self) -> float: ...


class PIDMotor(EncoderMotor):
    def set_target_position(self, pos: float): ...
    def set_target_velocity(self, vel: float): ...



from dataclasses import dataclass
from typing import Optional

import rev

from robotpy_toolkit_7407.motor import PIDMotor


@dataclass
class SparkMaxConfig:
    k_P: Optional[float] = None
    k_I: Optional[float] = None
    k_D: Optional[float] = None
    k_F: Optional[float] = None
    output_range: Optional[tuple[float, float]] = None
    idle_mode: Optional[rev.CANSparkMax.IdleMode] = None


class SparkMax(PIDMotor):
    _motor: rev.CANSparkMax
    __pid_controller: rev.SparkMaxPIDController
    __encoder: rev.SparkMaxRelativeEncoder

    def __init__(self, can_id: int, brushless: bool = True, config: SparkMaxConfig = None):
        super().__init__()
        self._can_id = can_id
        self._brushless = brushless
        self._config = config

    def init(self):
        self._motor = rev.CANSparkMax(
            self._can_id,
            rev.CANSparkMax.MotorType.kBrushless if self._brushless else rev.CANSparkMax.MotorType.kBrushed
        )
        self.__pid_controller = self._motor.getPIDController()
        self.__encoder = self._motor.getEncoder()
        self._set_config(self._config)

    def set_raw_output(self, x: float):
        self._motor.set(x)

    def set_target_position(self, pos: float):
        self.__pid_controller.setReference(pos, rev.CANSparkMax.ControlType.kPosition)

    def set_target_velocity(self, vel: float):
        self.__pid_controller.setReference(vel, rev.CANSparkMax.ControlType.kVelocity)

    def get_sensor_position(self) -> float:
        return self.__encoder.getPosition()

    def get_sensor_velocity(self) -> float:
        return self.__encoder.getVelocity()

    def _set_config(self, config: SparkMaxConfig):
        if config is None:
            return
        if config.k_P is not None:
            self.__pid_controller.setP(config.k_P)
        if config.k_I is not None:
            self.__pid_controller.setI(config.k_I)
        if config.k_D is not None:
            self.__pid_controller.setD(config.k_D)
        if config.k_F is not None:
            self.__pid_controller.setFF(config.k_F)
        if config.output_range is not None:
            self.__pid_controller.setOutputRange(config.output_range[0], config.output_range[1])
        if config.idle_mode is not None:
            self._motor.setIdleMode(config.idle_mode)



from robotpy_toolkit_7407.network.network_system import Network
from robotpy_toolkit_7407.network.objects import (
    NetworkObject,
    Sendable,
    SubsystemNetworkObject,
    MotorNetworkObject,
    RobotStatusPacket
)



import math

from robotpy_toolkit_7407.oi.joysticks import JoystickAxis
from robotpy_toolkit_7407.subsystem import Subsystem
from robotpy_toolkit_7407.utils import logger
from robotpy_toolkit_7407.utils.math import rotate_vector


class SwerveNode:
    motor_reversed: bool
    motor_flip_diff: float

    def init(self):
        self.motor_reversed = False
        self.motor_flip_diff = 0

    def set(self, vel_tw_per_second: float, angle_radians: float):
        self.set_angle_radians(angle_radians, self.get_current_angle_raw() + self.motor_flip_diff)
        self.set_velocity_raw(vel_tw_per_second)

    # pos=0 - facing right, counter-clockwise around, pos=1 - facing right
    def set_angle_raw(self, pos: float): ...

    def get_current_angle_raw(self) -> float: ...

    # 0 degrees is facing right
    def set_angle_radians(self, target_radians: float, initial_radians: float):
        # Actual angle difference in radians
        diff = math.fmod(target_radians, 2 * math.pi) - math.fmod(initial_radians, 2 * math.pi)
        while diff > math.pi:
            diff -= 2 * math.pi
        while diff < -math.pi:
            diff += 2 * math.pi

        # Should we flip the motor
        if abs(diff) > math.pi / 2:
            if diff < 0:
                self.motor_flip_diff += math.pi
                diff += math.pi
            else:
                self.motor_flip_diff -= math.pi
                diff -= math.pi
            self.motor_reversed = not self.motor_reversed

        # Add diff to theta f
        theta_f = initial_radians + diff

        self.set_angle_raw(theta_f - self.motor_flip_diff)

    def set_velocity_raw(self, vel_tw_per_second: float): ...

    def get_current_velocity(self) -> float: ...


class SwerveOdometry:
    def init(self): ...
    def get_robot_angle_degrees(self) -> float: ...
    def reset_angle(self): ...


class SwerveDrivetrain(Subsystem):
    n_00: SwerveNode  # Top Left
    n_01: SwerveNode  # Bottom Left
    n_10: SwerveNode  # Top Right
    n_11: SwerveNode  # Bottom Right
    odometry: SwerveOdometry
    axis_dx: JoystickAxis
    axis_dy: JoystickAxis
    axis_rotation: JoystickAxis
    track_width: float

    def init(self):
        logger.info("initializing swerve drivetrain", "[swerve_drivetrain]")
        self.n_00.init()
        self.n_01.init()
        self.n_10.init()
        self.n_11.init()
        self.odometry.init()
        logger.info("initialization complete", "[swerve_drivetrain]")

    def set(self, vel_tw_per_second: tuple[float, float], angular_vel: float):
        vel_tw_per_second = rotate_vector(
            vel_tw_per_second[0],
            vel_tw_per_second[1],
            -self.odometry.get_robot_angle_degrees() * (math.pi / 180)
        )

        # logger.info(f"ROBOT AT {self.odometry.get_robot_angle_degrees()} degrees offset")
        # logger.info(f"({vel_tw_per_second[0]} tw/sec, {vel_tw_per_second[1]} tw/sec, {angular_vel} rad/sec)")

        if abs(vel_tw_per_second[0]) < 0.1 and abs(vel_tw_per_second[1]) < 0.1 and abs(angular_vel) < 0.1:
            self.n_00.set_velocity_raw(0)
            self.n_01.set_velocity_raw(0)
            self.n_10.set_velocity_raw(0)
            self.n_11.set_velocity_raw(0)
            return

        self.n_00.set(*self._swerve_displacement(-1, -1, vel_tw_per_second[0], vel_tw_per_second[1], angular_vel, 0))
        self.n_01.set(*self._swerve_displacement(-1, 1, vel_tw_per_second[0], vel_tw_per_second[1], angular_vel, math.pi))
        self.n_10.set(*self._swerve_displacement(1, -1, vel_tw_per_second[0], vel_tw_per_second[1], angular_vel, math.pi))
        self.n_11.set(*self._swerve_displacement(1, 1, vel_tw_per_second[0], vel_tw_per_second[1], angular_vel, 0))

    def stop(self):
        self.n_00.set(0, 0)
        self.n_01.set(0, 0)
        self.n_10.set(0, 0)
        self.n_11.set(0, 0)

    @staticmethod
    def _swerve_displacement(node_x: float, node_y: float, dx: float, dy: float, d_theta: float, angle_offset: float) -> tuple[float, float]:
        tangent_x, tangent_y = -node_y, node_x
        tangent_m = math.sqrt(tangent_x*tangent_x + tangent_y*tangent_y)
        tangent_x /= tangent_m
        tangent_y /= tangent_m

        r = math.sqrt(2) / 2
        sx = dx + r * d_theta * tangent_x
        sy = dy + r * d_theta * tangent_y

        theta = math.atan2(sy, sx)
        magnitude = math.sqrt(sx * sx + sy * sy)
        return magnitude, theta + angle_offset



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['robotpy_toolkit_7407',
 'robotpy_toolkit_7407.motors',
 'robotpy_toolkit_7407.network',
 'robotpy_toolkit_7407.oi',
 'robotpy_toolkit_7407.sensors',
 'robotpy_toolkit_7407.sensors.gyro',
 'robotpy_toolkit_7407.subsystem_templates',
 'robotpy_toolkit_7407.subsystem_templates.drivetrain',
 'robotpy_toolkit_7407.utils']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.0,<2.0.0', 'robotpy[ctre,commands2,rev]>=2022.0.21,<2023.0.0']

setup_kwargs = {
    'name': 'robotpy-toolkit-7407',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Ethan Chapman',
    'author_email': 'ethan.chapman0@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import math

from robotpy_toolkit_7407.command import SubsystemCommand
from robotpy_toolkit_7407.subsystem_templates.drivetrain.swerve_drivetrain import SwerveDrivetrain


class DriveSwerve(SubsystemCommand[SwerveDrivetrain]):
    def initialize(self) -> None:
        pass

    def execute(self) -> None:
        dx, dy, d_theta = self.subsystem.axis_dx.value, self.subsystem.axis_dy.value, self.subsystem.axis_rotation.value

        if abs(dx) < 0.1:
            dx = 0
        if abs(dy) < 0.1:
            dy = 0
        if abs(d_theta) < 0.1:
            d_theta = 0

        dx *= 4
        dy *= -4
        d_theta *= -4 * math.pi / 3

        self.subsystem.set((dx, dy), d_theta)

    def end(self, interrupted: bool) -> None:
        self.subsystem.stop()

    def isFinished(self) -> bool:
        return False

    def runsWhenDisabled(self) -> bool:
        return False
